#include <ivt.h>

void setInterruptHandler(void (*handler)(void), int vector)
{
	uint32_t *ivt = (uint32_t *) InterruptVectorTableAddress;
	uint32_t offset = 0;

	offset = (((uint32_t) handler - (uint32_t) &(ivt[vector]) - 0x8) >> 2);

	if(offset & 0xff000000) {
		debug("[er] Interrupt handler address out of range\n\r");
		halt();
	}
	
	/* Branch & Link Instruction (BAL) */
	ivt[vector] = 0xea000000 | offset;
}

void C_undefinedInstruction(uint32_t *sp)
{
	debug("[ex] Undefined Instruction\n\r");
	debugStack(sp, 17);
	halt();
}

void C_preFetchAbort(uint32_t *sp)
{
	debug("[ex] Pre-fetch Abort\n\r");
	debug("[db] %x - Fault Address Register\n\r", getFaultAddress());
	debug("[db] %x - Instruction Fault Address Register\n\r", getInstructionFaultAddress());
	debugInstructionFaultStatus(getInstructionFaultStatus());
	debugStack(sp, 17);
	halt();
}

void C_dataAbort(uint32_t *sp)
{
	debug("[ex] Data Abort\n\r");
	debug("[db] %x - Fault Address Register\n\r", getFaultAddress());
	debugDataFaultStatus(getDataFaultStatus());
	debugStack(sp, 17);
	halt();
}

void C_irq()
{
	systemTimerIrq();

	uart_irq();

	if(scheduler_irq() || sleepIrq()) schedule();
}

void C_fiq()
{
	debug("[ex] Fast Interrupt Request\n\r");
}

void halt()
{
	debug("[db] System halt\n\r");
	for(;;);
}

/* Debug */

void debugStack(uint32_t *sp, int count)
{
	debug("[db] Stack\n\r");
	for(int i = 0; i < count; i++)
		debug("\t%x: %x (%d)\n\r", (uint32_t) &(sp[i]), sp[i], sp[i]);
}

const char *dataFaultStatusText0[] = {
	"No function",
	"Alignment fault",
	"Instruction debug event fault",
	"Access bit fault on section",
	"Instruction cache maintenance operation fault",
	"Translation section fault",
	"Access bit fault on page",
	"Translation page fault",
	"Precise external abort",
	"Domain section fault",
	"No function",
	"Domain page fault",
	"External abort on translation, first level",
	"Permission section fault",
	"External abort on translation, second level",
	"Permission page fault"
};

const char *dataFaultStatusText1[] = {
	"No function",
	"No function",
	"No function",
	"No function",
	"No function",
	"No function",
	"Imprecise external abort",
	"No function",
	"No function",
	"No function",
	"No function",
	"No function",
	"No function",
	"No function",
	"No function",
	"No function"
};

void debugDataFaultStatus(uint32_t value)
{

	uint32_t SD = (value >> 12) & 1;
	uint32_t isRW = (value >> 11) & 1;
	uint32_t S = (value >> 10) & 1;
	uint32_t status = value & 0xf;

	debug("[db] Data Fault Register\n\r");
	debug("\t@ %s error caused the abort\n\r", (SD)?"AXI slave":"AXI decode");
	debug("\t@ %s access caused the abort\n\r", (isRW)?"Write":"Read");
	if(S)
		debug("\t@ Status: %s\n\r", dataFaultStatusText1[status]);
	else
		debug("\t@ Status: %s\n\r", dataFaultStatusText0[status]);

}

const char *instructionFaultStatusText[] = {
	"No function",
	"Alignment fault",
	"Instruction debug event fault",
	"Access bit fault on section",
	"No function",
	"Translation section fault",
	"Access bit fault on page",
	"Translation page fault",
	"Precise external abort",
	"Domain section fault",
	"No function",
	"Domain page fault",
	"External abort on translation, first level",
	"Permission section fault",
	"External abort on translation, second level",
	"Permission page fault"
};

void debugInstructionFaultStatus(uint32_t value)
{

	uint32_t SD = (value >> 12) & 1;
	uint32_t status = value & 0xf;

	debug("[db] Instruction Fault Register\n\r");
	debug("\t@ %s error caused the abort\n\r", (SD)?"AXI slave":"AXI decode");
	debug("\t@ Status: %s\n\r", instructionFaultStatusText[status]);
}

const char *exceptions[] = {
	"Reset",
	"Undefined Instruction",
	"Software Interrupt",
	"Pre-fetch Abort",
	"Data Abort",
	"Reserved",
	"Interrupt Request",
	"Fast Interrupt Request"
};

void debugInterruptVectorTable(uint32_t *ivt)
{
	debug("[db] Interrupt Vector Table\n\r");
	for(int i = 0; i < 8; i++)
		debug("\t%x - %x (%s)\n\r", 
			(((ivt[i] & 0xffffff) << 2) + (uint32_t) &(ivt[i]) + 0x8), ivt[i], exceptions[i]);
}
